SOC_IRAM_LOW = 0x4037c000
SOC_IRAM_HIGH = 0x403c0000
SOC_DRAM_LOW = 0x3fca0000
SOC_DRAM_HIGH = 0x3fce0000
SOC_RTC_DATA_LOW = 0xffffffff
SOC_RTC_DATA_HIGH = 0xffffffff
SOC_RTC_DRAM_LOW = 0xffffffff
SOC_RTC_DRAM_HIGH = 0xffffffff
