SOC_IRAM_LOW = 0x4ff00000
SOC_IRAM_HIGH = 0x4ffc0000
SOC_DRAM_LOW = 0x4ff00000
SOC_DRAM_HIGH = 0x4ffc0000
SOC_RTC_IRAM_LOW = 0x50108000
SOC_RTC_IRAM_HIGH = 0x50110000
SOC_RTC_DRAM_LOW = 0x50108000
SOC_RTC_DRAM_HIGH = 0x50110000
SOC_RTC_DATA_LOW = 0x50108000
SOC_RTC_DATA_HIGH = 0x50110000
