SOC_IRAM_LOW = 0x40370000
SOC_IRAM_HIGH = 0x403e0000
SOC_DRAM_LOW = 0x3fc88000
SOC_DRAM_HIGH = 0x3fd00000
SOC_RTC_DRAM_LOW = 0x600fe000
SOC_RTC_DRAM_HIGH = 0x60100000
SOC_RTC_DATA_LOW = 0x50000000
SOC_RTC_DATA_HIGH = 0x50002000
