SOC_IRAM_LOW = 0x40800000
SOC_IRAM_HIGH = 0x40880000
SOC_DRAM_LOW = 0x40800000
SOC_DRAM_HIGH = 0x40880000
SOC_RTC_DRAM_LOW = 0x50000000
SOC_RTC_DRAM_HIGH = 0x50004000
SOC_RTC_DATA_LOW = 0x50000000
SOC_RTC_DATA_HIGH = 0x50004000
